@@ -6,13 +6,7 @@
 # from the MIT License (MIT).
 #
 
-"""Implements reading and writing to/from HDFS.
-
-The main entry point is the :func:`~smart_open.hdfs.open` function.
-
-Uses the command-line hdfs utility under the covers.
-
-"""
+"""Implements reading and writing to/from HDFS."""
 
 import io
 import logging
@@ -74,11 +68,6 @@ class CliRawInputBase(io.RawIOBase):
         self._uri = uri
         self._sub = subprocess.Popen(["hdfs", "dfs", '-cat', self._uri], stdout=subprocess.PIPE)
 
-        #
-        # This member is part of the io.BufferedIOBase interface.
-        #
-        self.raw = None
-
     #
     # Override some methods from io.IOBase.
     #
@@ -138,11 +127,6 @@ class CliRawOutputBase(io.RawIOBase):
         self._sub = subprocess.Popen(["hdfs", "dfs", '-put', '-f', '-', self._uri],
                                      stdin=subprocess.PIPE)
 
-        #
-        # This member is part of the io.RawIOBase interface.
-        #
-        self.raw = None
-
     def close(self):
         logger.debug("close: called")
         if not self.closed:
