@@ -104,15 +104,17 @@ class BufferedInputBase(io.BufferedIOBase):
                  kerberos=False, user=None, password=None, cert=None,
                  headers=None, session=None, timeout=None):
 
+        self.url = url
+        self.cert = cert
         self.session = session or requests
 
         if kerberos:
             import requests_kerberos
-            auth = requests_kerberos.HTTPKerberosAuth()
+            self.auth = requests_kerberos.HTTPKerberosAuth()
         elif user is not None and password is not None:
-            auth = (user, password)
+            self.auth = (user, password)
         else:
-            auth = None
+            self.auth = None
 
         self.buffer_size = buffer_size
         self.mode = mode
@@ -125,9 +127,9 @@ class BufferedInputBase(io.BufferedIOBase):
         self.timeout = timeout
 
         self.response = self.session.get(
-            url,
-            auth=auth,
-            cert=cert,
+            self.url,
+            auth=self.auth,
+            cert=self.cert,
             stream=True,
             headers=self.headers,
             timeout=self.timeout,
@@ -136,15 +138,9 @@ class BufferedInputBase(io.BufferedIOBase):
         if not self.response.ok:
             self.response.raise_for_status()
 
-        self._read_iter = self.response.iter_content(self.buffer_size)
         self._read_buffer = bytebuffer.ByteBuffer(buffer_size)
         self._current_pos = 0
 
-        #
-        # This member is part of the io.BufferedIOBase interface.
-        #
-        self.raw = None
-
     #
     # Override some methods from io.IOBase.
     #
@@ -153,7 +149,7 @@ class BufferedInputBase(io.BufferedIOBase):
         logger.debug("close: called")
         if not self.closed:
             self.response = None
-            self._read_iter = None
+            self._read_buffer = None
 
     @property
     def closed(self):
@@ -177,32 +173,23 @@ class BufferedInputBase(io.BufferedIOBase):
         """
         Mimics the read call to a filehandle object.
         """
+        if size < -1:
+            raise ValueError(f'size must be >= -1, got {size}')
+
         logger.debug("reading with size: %d", size)
-        if self.response is None:
-            return b''
+        if self.closed or size == 0:
+            return b""
 
-        if size == 0:
-            return b''
-        elif size < 0 and len(self._read_buffer) == 0:
-            retval = self.response.raw.read()
-        elif size < 0:
-            retval = self._read_buffer.read() + self.response.raw.read()
+        if size == -1:
+            if len(self._read_buffer):
+                retval = self._read_buffer.read() + self.response.raw.read()
+            else:  # Avoid unnecessary +
+                retval = self.response.raw.read()
         else:
+            # Fill _read_buffer until it contains enough bytes
             while len(self._read_buffer) < size:
-                logger.debug(
-                    "http reading more content at current_pos: %d with size: %d",
-                    self._current_pos, size,
-                )
-                bytes_read = self._read_buffer.fill(self._read_iter)
-                if bytes_read == 0:
-                    # Oops, ran out of data early.
-                    retval = self._read_buffer.read()
-                    self._current_pos += len(retval)
-
-                    return retval
-
-            # If we got here, it means we have enough data in the buffer
-            # to return to the caller.
+                if self._read_buffer.fill(self.response.raw) == 0:
+                    break  # EOF reached
             retval = self._read_buffer.read(size)
 
         self._current_pos += len(retval)
@@ -238,35 +225,7 @@ class SeekableBufferedInputBase(BufferedInputBase):
 
         If none of those are set, will connect unauthenticated.
         """
-        self.url = url
-
-        self.session = session or requests
-
-        if kerberos:
-            import requests_kerberos
-            self.auth = requests_kerberos.HTTPKerberosAuth()
-        elif user is not None and password is not None:
-            self.auth = (user, password)
-        else:
-            self.auth = None
-
-        if headers is None:
-            self.headers = _HEADERS.copy()
-        else:
-            self.headers = headers
-
-        self.cert = cert
-        self.timeout = timeout
-
-        self.buffer_size = buffer_size
-        self.mode = mode
-        self.response = self._partial_request()
-
-        if not self.response.ok:
-            self.response.raise_for_status()
-
-        logger.debug('self.response: %r, raw: %r', self.response, self.response.raw)
-
+        super().__init__(url, mode, buffer_size, kerberos, user, password, cert, headers, session, timeout)
         self.content_length = int(self.response.headers.get("Content-Length", -1))
         #
         # We assume the HTTP stream is seekable unless the server explicitly
@@ -276,15 +235,6 @@ class SeekableBufferedInputBase(BufferedInputBase):
         #
         self._seekable = self.response.headers.get("Accept-Ranges", "").lower() != "none"
 
-        self._read_iter = self.response.iter_content(self.buffer_size)
-        self._read_buffer = bytebuffer.ByteBuffer(buffer_size)
-        self._current_pos = 0
-
-        #
-        # This member is part of the io.BufferedIOBase interface.
-        #
-        self.raw = None
-
     def seek(self, offset, whence=0):
         """Seek to the specified position.
 
@@ -320,13 +270,11 @@ class SeekableBufferedInputBase(BufferedInputBase):
 
         if new_pos == self.content_length:
             self.response = None
-            self._read_iter = None
             self._read_buffer.empty()
         else:
             response = self._partial_request(new_pos)
             if response.ok:
                 self.response = response
-                self._read_iter = self.response.iter_content(self.buffer_size)
                 self._read_buffer.empty()
             else:
                 self.response = None
@@ -344,15 +292,16 @@ class SeekableBufferedInputBase(BufferedInputBase):
         raise io.UnsupportedOperation
 
     def _partial_request(self, start_pos=None):
+        headers = self.headers.copy()
         if start_pos is not None:
-            self.headers.update({"range": smart_open.utils.make_range_string(start_pos)})
+            headers["range"] = smart_open.utils.make_range_string(start_pos)
 
         response = self.session.get(
             self.url,
             auth=self.auth,
             stream=True,
             cert=self.cert,
-            headers=self.headers,
+            headers=headers,
             timeout=self.timeout,
         )
         return response
